# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Set, Tuple, List

from ..types import URI
from .stream import LiveStream, LiveStreamFactory
from .channel import LiveChannel
from .genre import LiveGenre


class LiveParser:

    def __init__(self):
        super().__init__()
        self.__factory = LiveStreamFactory()

    @property
    def stream_factory(self) -> LiveStreamFactory:
        return self.__factory

    def get_stream(self, url: URI) -> LiveStream:
        return self.stream_factory.get_stream(url=url)

    def parse(self, text: str) -> List[LiveGenre]:
        lines = text.splitlines()
        return self.parse_lines(lines=lines)

    # protected
    def parse_lines(self, lines: List[str]) -> List[LiveGenre]:
        all_groups: List[LiveGenre] = []
        current = LiveGenre(title='')
        for item in lines:
            text = item.strip()
            if len(text) == 0:
                continue
            elif text.startswith(r'#'):
                continue
            elif text.startswith(r'//'):
                continue
            # 1. check group name
            title = fetch_genre(text=text)
            if title is not None:
                # add current group
                if len(current.channels) > 0:
                    all_groups.append(current)
                # create next group
                current = LiveGenre(title=title)
                continue
            # 2. parse channel
            name, sources = fetch_channel(text=text)
            if name is None:
                # empty line?
                continue
            else:
                channel = LiveChannel(name=name)
            # 3. create streams
            streams: Set[LiveStream] = set()
            for src in sources:
                streams.add(self.get_stream(url=src))
            channel.add_streams(streams=streams)
            current.add_channel(channel=channel)
        # add last group
        if len(current.channels) > 0:
            all_groups.append(current)
        return all_groups


def fetch_genre(text: str) -> Optional[str]:
    """ get group title """
    pos = text.find(',#genre#')
    if pos >= 0:
        return text[:pos].strip()


def fetch_channel(text: str) -> Tuple[Optional[str], List[URI]]:
    """ get channel name & stream sources """
    pos = text.find(',http')
    if pos < 0:
        # not a channel line
        return None, []
    else:
        # fetch channel name
        name = text[:pos].strip()
        pos += 1  # skip ','
        text = text[pos:]
    # fetch sources
    return name, fetch_streams(text=text)


def fetch_streams(text: str) -> List[URI]:
    """ split steam sources with '#' """
    sources = []
    while True:
        pos = text.find('#http')
        if pos > 0:
            sources.append(text[:pos])
            pos += 1  # skip '#'
            text = text[pos:]
            continue
        # remove the tail
        pos = text.find('$')
        if pos > 0:
            text = text[:pos]
        else:
            text = text.rstrip()
        # last url
        sources.append(text)
        break
    return sources



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, List, Dict

import requests
from aiou import TextFile, JSONFile

from .types import URI
from .utils import Log, Logging

from .lives import LiveStream, LiveChannel, LiveGenre

from .config import LiveConfig
from .scanner import ScanContext, ScanEventHandler


class SourceLoader:

    def __init__(self):
        super().__init__()
        self.__resources: Dict[str, str] = {}

    async def load_text(self, src: Union[str, URI]) -> Optional[str]:
        # 1. check caches
        res = self.__resources.get(src)
        if res is None:
            # 2. cache not found, try to load
            res = await _load(src=src)
            if res is None:
                res = ''  # place holder
            # 3. cache the result
            self.__resources[src] = res
        # OK
        if len(res) > 0:
            return res


async def _load(src: Union[str, URI]) -> Optional[str]:
    if src.find(r'://') > 0:
        return await _http_get(url=src)
    else:
        return await _file_read(path=src)


async def _file_read(path: str) -> Optional[str]:
    try:
        return await TextFile(path=path).read()
    except Exception as error:
        Log.error(msg='failed to read file: %s, error: %s' % (path, error))


async def _http_get(url: URI) -> Optional[str]:
    try:
        response = requests.get(url)
        encoding = response.encoding
        if encoding is None or len(encoding) == 0:
            response.encoding = 'utf-8'
        elif encoding.upper() == 'ISO-8859-1':
            response.encoding = 'utf-8'
        return response.text
    except Exception as error:
        Log.error(msg='failed to get URL: %s, error: %s' % (url, error))


class LiveHandler(ScanEventHandler, Logging):
    """ Live Storage """

    def __init__(self, config: LiveConfig):
        super().__init__()
        self.__config = config

    @property
    def config(self) -> LiveConfig:
        return self.__config

    #
    #   Storage
    #

    async def update_index(self, lives: List[Dict]) -> bool:
        """ update 'tvbox.json' """
        path = self.config.get_output_index_path()
        if path is None or len(lives) == 0:
            self.warning(msg='ignore live index: %s => %s' % (path, lives))
            return False
        self.info(msg='saving index file: %d live urls -> %s' % (len(lives), path))
        return await JSONFile(path=path).write(container={
            'lives': lives,
        })

    async def update_lives(self, genres: List[LiveGenre], url: URI) -> bool:
        path = self.config.get_output_lives_path(url=url)
        if path is None or len(genres) == 0:
            self.warning(msg='ignore live urls: %s => %s' % (url, genres))
            return False
        count = 0
        text = ''
        for group in genres:
            array: List[str] = []
            channels = group.channels
            for item in channels:
                # get valid stream sources
                streams = item.streams
                sources: List[URI] = [src.url for src in streams if src.available]
                if len(sources) == 0:
                    self.warning(msg='empty channel: "%s" -> %s' % (item.name, streams))
                    continue
                line = '#'.join(sources)
                line = '%s,%s' % (item.name, line)
                array.append(line)
                count += 1
            if len(array) == 0:
                self.warning(msg='empty genre: "%s" -> %s' % (group.title, channels))
                continue
            text += '\n%s,#genre#\n' % group.title
            text += '\n%s\n' % '\n'.join(array)
        # OK
        self.info(msg='saving lives file: %d genres, %d channels -> %s' % (len(genres), count, path))
        if len(text) > 0:
            return await TextFile(path=path).write(text=text)

    async def update_source(self, text: str, url: URI) -> bool:
        path = self.config.get_output_source_path(url=url)
        if path is None:
            self.warning(msg='ignore index file: %s => %s' % (url, text))
            return False
        self.info(msg='saving source file: %d lines -> %s' % (len(text.splitlines()), path))
        if len(text) > 0:
            return await TextFile(path=path).write(text=text)

    #
    #   ScanEventHandler
    #

    # Override
    async def on_scan_start(self, context: ScanContext):
        pass

    # Override
    async def on_scan_finished(self, context: ScanContext):
        pass

    # genre events

    # Override
    async def on_scan_genre_start(self, context: ScanContext, genre: LiveGenre):
        pass

    # Override
    async def on_scan_genre_finished(self, context: ScanContext, genre: LiveGenre):
        pass

    # channel events

    # Override
    async def on_scan_channel_start(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        pass

    # Override
    async def on_scan_channel_finished(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        pass

    # stream events

    # Override
    async def on_scan_stream_start(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        pass

    # Override
    async def on_scan_stream_finished(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        offset = context.get_value(key='stream_offset', default=0)
        total = context.get_value(key='total_streams', default=0)
        self.info(msg='scanned (%d/%d) stream: "%s"\t-> %s' % (offset, total, channel.name, stream))



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Tuple, List, Dict

from .lives import LiveStream, LiveChannel, LiveGenre
from .lives import LiveStreamScanner
from .lives import LiveParser


class ScanContext:

    def __init__(self):
        super().__init__()
        self.__vars: Dict[str, Any] = {}

    def get_value(self, key: str, default: Any = None) -> Optional[Any]:
        return self.__vars.get(key, default)

    def set_value(self, key: str, value: Optional[Any]):
        if value is None:
            self.__vars.pop(key, None)
        else:
            self.__vars[key] = value


class ScanEventHandler(ABC):
    """ Scan callback """

    @abstractmethod
    async def on_scan_start(self, context: ScanContext):
        raise NotImplemented

    @abstractmethod
    async def on_scan_finished(self, context: ScanContext):
        raise NotImplemented

    # genre events

    @abstractmethod
    async def on_scan_genre_start(self, context: ScanContext, genre: LiveGenre):
        raise NotImplemented

    @abstractmethod
    async def on_scan_genre_finished(self, context: ScanContext, genre: LiveGenre):
        raise NotImplemented

    # channel events

    @abstractmethod
    async def on_scan_channel_start(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        raise NotImplemented

    @abstractmethod
    async def on_scan_channel_finished(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        raise NotImplemented

    # stream events

    @abstractmethod
    async def on_scan_stream_start(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        raise NotImplemented

    @abstractmethod
    async def on_scan_stream_finished(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        raise NotImplemented


class LiveScanner:

    def __init__(self):
        super().__init__()
        self.__parser = LiveParser()
        self.__scanner = LiveStreamScanner()

    @property  # protected
    def live_parser(self) -> LiveParser:
        return self.__parser

    @property  # protected
    def stream_scanner(self) -> LiveStreamScanner:
        return self.__scanner

    async def scan(self, text: str, timeout: float, handler: ScanEventHandler) -> List[LiveGenre]:
        """ Get non-empty channel groups """
        groups: List[LiveGenre] = []
        genres = self.live_parser.parse(text=text)
        total_genres = len(genres)
        total_channels, total_streams = _count_channel_streams(genres=genres)
        # prepare scan context
        context = ScanContext()
        context.set_value(key='timeout', value=timeout)
        context.set_value(key='all_genres', value=genres)
        context.set_value(key='available_genres', value=groups)
        context.set_value(key='total_genres', value=total_genres)
        context.set_value(key='total_channels', value=total_channels)
        context.set_value(key='total_streams', value=total_streams)
        context.set_value(key='genre_offset', value=0)
        context.set_value(key='channel_offset', value=0)
        context.set_value(key='stream_offset', value=0)
        # start mission
        await handler.on_scan_start(context=context)
        context.set_value(key='genre_count', value=total_genres)
        # positions
        offset = 0  # context.get_value(key='genre_offset', default=0)
        index = 0
        for item in genres:
            context.set_value(key='genre_index', value=index)
            await handler.on_scan_genre_start(context=context, genre=item)
            # scan genre start
            channels = await self._scan_genre(genre=item, context=context, handler=handler)
            if len(channels) > 0:
                item.channels = channels
                groups.append(item)
            # scan genre finished
            await handler.on_scan_genre_finished(context=context, genre=item)
            # increase positions
            index += 1
            offset += 1
            context.set_value(key='genre_offset', value=offset)
        # mission accomplished
        await handler.on_scan_finished(context=context)
        return groups

    async def _scan_genre(self, genre: LiveGenre,
                          context: ScanContext, handler: ScanEventHandler) -> List[LiveChannel]:
        """ Get available channels in this genre """
        available_channels: List[LiveChannel] = []
        channels = genre.channels
        context.set_value(key='channel_count', value=len(channels))
        # positions
        offset = context.get_value(key='channel_offset', default=0)
        index = 0
        for item in channels:
            context.set_value(key='channel_index', value=index)
            await handler.on_scan_channel_start(context=context, genre=genre, channel=item)
            # scan channel start
            streams = await self._scan_channel(channel=item, context=context, handler=handler)
            if len(streams) > 0:
                item.streams = streams
                available_channels.append(item)
            # scan channel finished
            await handler.on_scan_channel_finished(context=context, genre=genre, channel=item)
            # increase positions
            index += 1
            offset += 1
            context.set_value(key='channel_offset', value=offset)
        return available_channels

    async def _scan_channel(self, channel: LiveChannel,
                            context: ScanContext, handler: ScanEventHandler) -> List[LiveStream]:
        """ Get available streams in this channel """
        available_streams: List[LiveStream] = []
        streams = channel.streams
        context.set_value(key='stream_count', value=len(streams))
        # positions
        offset = context.get_value(key='stream_offset', default=0)
        index = 0
        for item in streams:
            context.set_value(key='stream_index', value=index)
            await handler.on_scan_stream_start(context=context, channel=channel, stream=item)
            # scan stream start
            timeout = context.get_value(key='timeout', default=None)
            src = await self.stream_scanner.scan_stream(stream=item, timeout=timeout)
            if src is not None:  # and src.available:
                available_streams.append(src)
            # scan stream finished
            await handler.on_scan_stream_finished(context=context, channel=channel, stream=item)
            # increase positions
            index += 1
            offset += 1
            context.set_value(key='stream_offset', value=offset)
        return available_streams


def _count_channel_streams(genres: List[LiveGenre]) -> Tuple[int, int]:
    total_channels = 0
    total_streams = 0
    for group in genres:
        channels = group.channels
        for item in channels:
            total_channels += 1
            total_streams += len(item.streams)
    return total_channels, total_streams



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
from typing import Optional, List, Dict

import requests

from ..types import URI
from ..utils import DateTime
from ..utils import AsyncRunner as Runner
from .stream import LiveStream


class LiveStreamScanner(Runner):

    def __init__(self):
        super().__init__(interval=2)
        # caches
        self.__checkers: Dict[URI, LiveStreamChecker] = {}
        self.__queue: List[LiveStreamChecker] = []
        self.__lock = threading.Lock()
        # background thread
        thr = Runner.async_thread(coro=self.start())
        thr.start()

    def _get_checker(self, stream: LiveStream):
        with self.__lock:
            url = stream.url
            checker = self.__checkers.get(url)
            if checker is None:
                checker = LiveStreamChecker(stream=stream)
                self.__checkers[url] = checker
            return checker

    async def scan_stream(self, stream: LiveStream, timeout: float = None) -> Optional[LiveStream]:
        """
        Check whether the stream source is available

            1. when it's available and not expired,
               return itself immediately;
            2. when it's checked before, add an update task in background thread,
               and return the stream itself immediately;
            3. otherwise, check the url, if respond nothing/error,
               return None, else return the stream.

        :param stream: live stream
        :param timeout:
        :return: None on invalid
        """
        checker = self._get_checker(stream=stream)
        with self.__lock:
            if checker.is_checked():
                # this stream was checked recently,
                # no need to check again.
                is_first = False
            elif stream.time is not None:
                # this stream was checked before, but expired now; so
                # add it in the waiting queue to check it in background thread.
                self.__queue.append(checker)
                is_first = False
            else:
                # first check
                is_first = True
        if is_first:
            await checker.check(timeout=timeout)
        if stream.available:
            return stream

    # Override
    async def process(self) -> bool:
        with self.__lock:
            if len(self.__queue) == 0:
                # nothing to do now
                return False
            checker = self.__queue.pop(0)
        try:
            await checker.check(timeout=64)
        except Exception as error:
            print('[TVBox] failed to scan stream: %s, error: %s' % (checker.stream, error))


class LiveStreamChecker:

    SCAN_INTERVAL = 3600 * 2

    def __init__(self, stream: LiveStream):
        super().__init__()
        self.__stream = stream
        self.__lock = threading.Lock()

    @property
    def stream(self) -> LiveStream:
        return self.__stream

    def is_checked(self, now: float = None) -> bool:
        """ whether checked recently """
        last_time = self.stream.time
        if last_time is None:
            # never checked yet
            return False
        elif now is None:
            now = DateTime.current_timestamp()
        # check interval
        return now < (last_time + self.SCAN_INTERVAL)

    async def check(self, timeout: float = None) -> bool:
        stream = self.stream
        with self.__lock:
            if self.is_checked():
                # no need to check again now
                return True
            # check it again
            ttl = await check_stream(stream=stream, timeout=timeout)
            return ttl is not None and ttl > 0


async def check_stream(stream: LiveStream, timeout: float = None) -> Optional[float]:
    start_time = DateTime.current_timestamp()
    available = await _http_check(url=stream.url, timeout=timeout)
    end_time = DateTime.current_timestamp()
    if available:
        ttl = end_time - start_time
    else:
        ttl = None
    stream.set_ttl(ttl=ttl, now=end_time)
    return ttl


async def _http_check(url: URI, timeout: float = None, is_m3u8: bool = None) -> bool:
    if is_m3u8 is None and url.lower().find('m3u8') > 0:
        is_m3u8 = True
    try:
        response = requests.head(url, timeout=timeout)
        status_code = response.status_code
        if status_code == 302:
            redirected_url = response.headers.get('Location')
            return await _http_check(url=redirected_url, is_m3u8=is_m3u8, timeout=timeout)
        elif status_code != 200:
            # HTTP error
            return False
        elif is_m3u8:
            return True
        # check content type
        content_type = response.headers.get('Content-Type')
        content_type = '' if content_type is None else str(content_type).lower()
        if content_type.find('application/vnd.apple.mpegurl') >= 0:
            return True
        elif content_type.find('application/x-mpegurl') >= 0:
            return True
    except Exception as error:
        print('[TVBox] failed to query URL: %s, error: %s' % (url, error))
    return False



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Any, List, Dict

from aiou import Path

from aiou import TextFile

from .types import URI, MapInfo
from .utils import Log
from .utils import parse_json
from .utils import hex_md5


class Config(MapInfo):

    def get_option(self, section: str, option: str) -> Optional[Any]:
        sub: Dict = self.get(key=section, default={})
        return sub.get(option)

    @classmethod
    async def load(cls, path: str):
        info = await _file_read(path=path)
        if info is not None:
            info = parse_json(text=info)
            return cls(info=info)


class LiveConfig(Config):

    @property
    def sources(self) -> List[URI]:
        array = self.get_option(section='tvbox', option='sources')
        return [] if array is None else array

    @property
    def lives(self) -> List[URI]:
        array = self.get_option(section='tvbox', option='lives')
        return [] if array is None else array

    @property
    def base_url(self) -> Optional[URI]:
        """ base url """
        return self.get_option(section='tvbox', option='base-url')

    @property
    def output_dir(self) -> Optional[str]:
        """ output directory """
        return self.get_option(section='tvbox', option='output-dir')

    def get_index_file(self) -> Optional[str]:
        return self.get_option(section='tvbox', option='index-file')

    def get_lives_file(self, url: URI) -> Optional[str]:
        file = self.get_option(section='tvbox', option='lives-file')
        if file is not None:
            digest = hex_md5(data=url)
            return file.replace(r'{HASH}', digest)

    def get_source_file(self, url: URI) -> Optional[str]:
        file = self.get_option(section='tvbox', option='source-file')
        if file is not None:
            digest = hex_md5(data=url)
            return file.replace(r'{HASH}', digest)

    #
    #   Output file path
    #

    def get_output_index_path(self) -> str:
        """ output index """
        file = self.get_index_file()
        return join_path(base=self.output_dir, file=file)

    def get_output_lives_path(self, url: URI) -> str:
        """ output lives """
        file = self.get_lives_file(url=url)
        return join_path(base=self.output_dir, file=file)

    def get_output_source_path(self, url: URI) -> str:
        file = self.get_source_file(url=url)
        return join_path(base=self.output_dir, file=file)

    #
    #   Output URL
    #

    def get_output_index_url(self) -> Optional[URI]:
        file = self.get_index_file()
        return join_path(base=self.base_url, file=file)

    def get_output_lives_url(self, url: URI):
        file = self.get_lives_file(url=url)
        return join_path(base=self.base_url, file=file)


def join_path(base: str, file: str) -> str:
    assert base is not None and len(base) > 0, 'base path empty, file: %s' % file
    assert file is not None and len(file) > 0, 'file name empty, path: %s' % base
    # standard base path
    if base.endswith(r'/'):
        base = base.rstrip(r'/')
    else:
        base = Path.dir(path=base)
    # standard file name
    if file.startswith(r'./'):
        file = file[2:]
    return '%s/%s' % (base, file)


async def _file_read(path: str) -> Optional[str]:
    try:
        return await TextFile(path=path).read()
    except Exception as error:
        Log.error(msg='failed to read file: %s, error: %s' % (path, error))
